/**
 * @file logging.c Debug logging support
 *
 * Copyright (C) 2004-2005 Christian Hammond.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA  02111-1307  USA
 */
#include <libgalago/galago.h>
#include <stdio.h>
#include "logging.h"

static char *_filename = NULL;
static gboolean _verbose = FALSE;
static FILE *_fp = NULL;

static const char *level_names[] =
{
	NULL, "MISC", "INFO", "WARNING", "ERROR", "FATAL", "ASSERT"
};

static void
log_handler(GalagoLogLevel level, const char *msg, va_list args,
			void *user_data)
{
	if (_fp != NULL)
	{
		va_list tmp;

		fprintf(_fp, "%s: ", level_names[level]);
		G_VA_COPY(tmp, args);
		vfprintf(_fp, msg, tmp);
		va_end(tmp);

		fflush(_fp);
	}

	if (_verbose)
	{
		printf("galagod:%s: ", level_names[level]);
		vprintf(msg, args);
	}
}

void
galagod_logging_set_filename(const char *filename)
{
	galago_return_if_fail(*filename != '\0');

	if (_filename != NULL)
		g_free(_filename);

	_filename = (filename != NULL ? g_strdup(filename) : NULL);
}

void
galagod_logging_set_verbose(gboolean verbose)
{
	_verbose = verbose;
}

void
galagod_logging_start(void)
{
	if (_filename != NULL)
	{
		if ((_fp = fopen(_filename, "w")) == NULL)
			galago_log_fatal("Can't open log file %s for writing\n", _filename);
	}

	galago_log_set_handler(GALAGO_LOG_ALL, log_handler, NULL);
}

void
galagod_logging_stop(void)
{
	if (_fp != NULL)
	{
		fclose(_fp);
		_fp = NULL;
	}

	if (_filename != NULL)
	{
		g_free(_filename);
		_filename = NULL;
	}

	galago_log_unset_handler(GALAGO_LOG_ALL);
}
