/**
 * @file meta-presence.c Meta-Presence API
 *
 * Copyright (C) 2004-2005 Christian Hammond.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA  02111-1307  USA
 */
#include "meta-presence.h"
#include "coco.h"
#include "filter.h"

static void
_galagod_dbus_presence_emit_idle_updated(GalagoPresence *presence,
										 const char *unused,
										 gboolean idle,
										 dbus_uint32_t idle_time)
{
	DBusMessage *message;
	DBusMessageIter iter;

	galago_return_if_fail(presence != NULL);

	message = dbus_message_new_signal(galago_object_get_dbus_path(presence),
									  GALAGO_DBUS_PRESENCE_INTERFACE,
									  "IdleUpdated");

	dbus_message_iter_init_append(message, &iter);
	galago_dbus_message_iter_append_boolean(&iter, idle);
	galago_dbus_message_iter_append_uint32(&iter, idle_time);

	dbus_connection_send(galago_core_get_dbus_conn(), message, NULL);
	dbus_message_unref(message);
}

GalagodMetaPresence *
galagod_meta_presence_new(GalagodMetaAccount *meta_account)
{
	GalagodMetaPresence *meta_presence;
	GalagoAccount *main_account;

	galago_return_val_if_fail(meta_account != NULL, NULL);

	meta_presence = g_new0(GalagodMetaPresence, 1);

	main_account = galagod_meta_account_get_account(meta_account);

	meta_presence->presence = galago_presence_new(main_account);

	galago_object_set_data(meta_presence->presence, "meta-presence",
						   meta_presence);

	galago_signal_connect(meta_presence->presence, "idle-changed",
		GALAGO_CALLBACK(_galagod_dbus_presence_emit_idle_updated), NULL);

	galagod_presence_add_filter(meta_presence->presence);

	return meta_presence;
}

void
galagod_meta_presence_destroy(GalagodMetaPresence *meta_presence)
{
	galago_return_if_fail(meta_presence != NULL);

	while (meta_presence->presences != NULL)
	{
		galagod_meta_presence_remove_presence(meta_presence,
			(GalagoPresence *)meta_presence->presences->data);
	}

	galagod_presence_remove_filter(meta_presence->presence);
	galago_object_unref(meta_presence->presence);

	g_free(meta_presence);
}

GalagoPresence *
galagod_meta_presence_get_presence(const GalagodMetaPresence *meta_presence)
{
	galago_return_val_if_fail(meta_presence != NULL, NULL);

	return meta_presence->presence;
}

const GalagoList *
galagod_meta_presence_get_presences(const GalagodMetaPresence *meta_presence)
{
	galago_return_val_if_fail(meta_presence != NULL, NULL);

	return meta_presence->presences;
}

void
galagod_meta_presence_add_presence(GalagodMetaPresence *meta_presence,
								   GalagoPresence *presence)
{
	GalagoPresence *main_presence;
	const GalagoList *l;

	galago_return_if_fail(meta_presence != NULL);
	galago_return_if_fail(presence      != NULL);

	meta_presence->presences = galago_list_append(meta_presence->presences,
												  presence);

	galago_object_set_data(presence, "meta-presence", meta_presence);

	main_presence = galagod_meta_presence_get_presence(meta_presence);

	galago_presence_set_idle(main_presence,
							 galago_presence_is_idle(presence),
							 galago_presence_get_idle_time(presence));

	galago_context_push(galagod_get_context());

	for (l = galago_presence_get_statuses(presence); l != NULL; l = l->next)
	{
		GalagoStatus *status = (GalagoStatus *)l->data;
		GalagoStatus *new_status;

		new_status = galago_status_duplicate(status);

		galago_presence_add_status(main_presence, new_status);
	}

	galago_context_pop();

	galagod_presence_add_filter(presence);
}

void
galagod_meta_presence_remove_presence(GalagodMetaPresence *meta_presence,
									  GalagoPresence *presence)
{
	galago_return_if_fail(meta_presence != NULL);
	galago_return_if_fail(presence      != NULL);

	galagod_presence_remove_filter(presence);

	galago_object_set_data(presence, "meta-presence", NULL);

	meta_presence->presences = galago_list_remove(meta_presence->presences,
												  presence);
}
