/**
 * @file person-db.c Galago Person Database
 *
 * Copyright (C) 2004-2005 Christian Hammond.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA  02111-1307  USA
 */
#include <libgalago/galago.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <glib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>

#include "meta-person.h"
#include "meta-account.h"
#include "person-list.h"

typedef struct
{
	char *username;
	char *service_id;

} PersonDbKey;

#define DB_FORMAT_VERSION 1
#define FLUSH_TIMEOUT (1000 * 60 * 10) /* 10 minutes */

static FILE *db_fp = NULL;
static guint32 db_file_timeout = 0;
static GHashTable *db_table = NULL;

/* Thanks to gconfd for some of this. */
#ifndef HAVE_FLOCKFILE
# define flockfile(f) (void)1
# define funlockfile(f) (void)1
# define getc_unlocked(f) getc(f)
#endif /* !HAVE_FLOCKFILE */

static char *
get_data_dir(void)
{
	const char *xdg_data_home = g_getenv("XDG_DATA_HOME");

	if (xdg_data_home != NULL)
		return g_strdup(xdg_data_home);

	return g_build_path(G_DIR_SEPARATOR_S, g_get_home_dir(),
						".local", "share", "galago", NULL);
}

static void
ensure_dir_exists(const char *dir)
{
	char *parent;

	if (g_file_test(dir, G_FILE_TEST_EXISTS))
		return;

	parent = g_path_get_dirname(dir);

	if (parent == NULL)
		return;

	ensure_dir_exists(parent);

	g_free(parent);

	mkdir(dir, 0700);
}

static void
get_db_paths(char **dir, char **db_path)
{
	char *data_dir = get_data_dir();

	if (dir != NULL)
	{
		*dir = g_build_path(G_DIR_SEPARATOR_S, data_dir, NULL);
	}

	if (db_path != NULL)
	{
		*db_path = g_build_path(G_DIR_SEPARATOR_S, data_dir, "people.db", NULL);
	}

	g_free(data_dir);
}

static char *
read_line(FILE *fp)
{
	int c;
	GString *str;

	str = g_string_new(NULL);

	flockfile(fp);

	for (;;)
	{
		c = getc_unlocked(fp);

		switch (c)
		{
			case EOF:
				if (ferror(fp))
				{
					galago_log_error("Error reading from file: %s\n",
									 g_strerror(errno));
				}

			case '\n':
				goto finished;
				break;

			default:
				g_string_append_c(str, c);
				break;
		}
	}

finished:
	funlockfile(fp);

	if (str->len == 0)
	{
		g_string_free(str, TRUE);

		return NULL;
	}

	return g_string_free(str, FALSE);
}

#define SKIP_WHITESPACE(c) \
	for (; *(c) != '\0' && g_ascii_isspace(*(c)); (c)++) ;

static gboolean
parse_line(const char *line, gboolean *ret_empty, char **ret_uid,
		   char ***ret_params, size_t *ret_param_count)
{
	const char *c;
	GString *param_str;
	GPtrArray *param_array;
	gboolean syntax_error = FALSE;
	gboolean in_quotes = FALSE;

	*ret_empty       = FALSE;
	*ret_uid         = NULL;
	*ret_params      = NULL;
	*ret_param_count = 0;

	c = line;

	SKIP_WHITESPACE(c);

	if (*c == '#' || *c == '\0')
	{
		*ret_empty = TRUE;

		return TRUE;
	}

	if ((c = strchr(line, ' ')) == NULL)
	{
		*ret_uid = g_strdup(line);

		return TRUE;
	}

	*ret_uid = g_strndup(line, c - line);

	param_array = g_ptr_array_new();
	param_str   = g_string_new(NULL);

	SKIP_WHITESPACE(c);

	for (; *c != '\0' && !syntax_error; c++)
	{
		if (*c == '"')
		{
			if (in_quotes)
			{
				char *str;

				str = g_string_free(param_str, FALSE);
				g_ptr_array_add(param_array, str);

				param_str = g_string_new(NULL);

				c++;
				SKIP_WHITESPACE(c);

				if (*c == ',')
				{
					c++;
					SKIP_WHITESPACE(c);
					c--;
				}
				else if (*c == '\0')
				{
					break;
				}
				else
				{
					syntax_error = TRUE;
					break;
				}
			}

			in_quotes = !in_quotes;
		}
		else if (*c == '\\')
		{
			if (!in_quotes)
			{
				syntax_error = TRUE;
				break;
			}

			switch (*++c)
			{
				case '\\':
				case '"':
					g_string_append_c(param_str, *c);
					break;

				case 'n':
					g_string_append_c(param_str, '\n');
					break;

				case 't':
					g_string_append_c(param_str, '\t');
					break;

				default:
					galago_log_warning("Invalid escaped character \\%c\n", *c);
					return FALSE;
					break;
			}
		}
		else if (!in_quotes)
		{
			if (!g_ascii_isspace(*c))
				syntax_error = TRUE;
		}
		else
		{
			g_string_append_c(param_str, *c);
		}
	}

	if (syntax_error)
	{
		g_ptr_array_free(param_array, TRUE);

		g_free(*ret_uid);
		*ret_uid = NULL;

		return FALSE;
	}

	g_ptr_array_add(param_array, NULL);

	*ret_param_count = param_array->len - 1;
	*ret_params = (char **)g_ptr_array_free(param_array, FALSE);

	return TRUE;
}

static void
process_line(const char *line, size_t line_num)
{
	gboolean empty, valid;
	char *uid;
	char **params;
	size_t i, param_count;

	valid = parse_line(line, &empty, &uid, &params, &param_count);

	if (empty)
	{
		return;
	}
	else if (!valid)
	{
		galago_log_warning("Syntax error in person.db on line %zu\n",
						   line_num);

		return;
	}

	for (i = 0; i < param_count; i += 2)
	{
		PersonDbKey *key;

		key = g_new0(PersonDbKey, 1);
		key->service_id = g_strdup(params[i]);
		key->username   = g_strdup(params[i + 1]);

		g_hash_table_insert(db_table, key, g_strdup(uid));

		galago_person_new(uid, TRUE); /* XXX */
	}

	g_free(uid);
	g_free(params);
}

static void
load_db(void)
{
	char *line;
	size_t line_num = 0;
	int ver;

	line = read_line(db_fp);

	if (line == NULL || *line == '\0')
		return;

	sscanf(line, "%d", &ver);

	if (ver != DB_FORMAT_VERSION)
	{
		galago_log_warning("Outdated or unknown person.db format version. "
						   "Ignoring file.\n");
	}
	else
	{
		while ((line = read_line(db_fp)) != NULL)
			process_line(line, ++line_num);
	}
}

const char *
galagod_person_db_find_with_account(GalagoAccount *account)
{
	GalagoService *service;
	PersonDbKey key;
	const char *id;

	galago_return_val_if_fail(account != NULL, NULL);

	service = galago_account_get_service(account);

	key.service_id = (char *)galago_service_get_id(service);
	key.username   = (char *)galago_account_get_username(account);

	id = g_hash_table_lookup(db_table, &key);

	return id;
}

void
galagod_person_db_flush(void)
{
	const GalagoList *l, *l2, *accounts;
	GalagodMetaPerson *meta_person;
	GalagoPerson *main_person;
	GalagoService *service;
	GalagoAccount *account;
	char *db_path;

	if (db_fp != NULL)
		fclose(db_fp);

	get_db_paths(NULL, &db_path);

	db_fp = fopen(db_path, "w");

	g_free(db_path);

	fprintf(db_fp, "%d\n", DB_FORMAT_VERSION);

	for (l = galagod_people_get_meta_people(); l != NULL; l = l->next)
	{
		gboolean first = TRUE;

		meta_person = (GalagodMetaPerson *)l->data;
		main_person = galagod_meta_person_get_person(meta_person);

		accounts = galago_person_get_accounts(main_person, FALSE);

		if (accounts == NULL)
			continue;

		fprintf(db_fp, "%s ", galago_person_get_id(main_person));

		for (l2 = accounts; l2 != NULL; l2 = l2->next)
		{
			account = (GalagoAccount *)l2->data;

			service = galago_account_get_service(account);

			fprintf(db_fp, "%s\"%s\", \"%s\"",
					(first ? "" : ", "),
					galago_service_get_id(service),
					galago_account_get_username(account));

			first = FALSE;
		}

		fputc('\n', db_fp);
	}

	fflush(db_fp);
}

static void
destroy_db_key(PersonDbKey *key)
{
	if (key->username != NULL)
		g_free(key->username);

	if (key->service_id != NULL)
		g_free(key->service_id);

	g_free(key);
}

static gboolean
timeout_cb(gpointer data)
{
	galagod_person_db_flush();

	return TRUE;
}

void
galagod_person_db_init(void)
{
	char *db_dir, *db_path;

	get_db_paths(&db_dir, &db_path);

	ensure_dir_exists(db_dir);

	if (!g_file_test(db_dir, G_FILE_TEST_EXISTS))
		mkdir(db_dir, 0700);

	db_table = g_hash_table_new_full(g_str_hash, g_str_equal,
										(GFreeFunc)destroy_db_key, g_free);

	if ((db_fp = fopen(db_path, "a+")) == NULL)
	{
		galago_log_error("Unable to open person db file %s for "
						 "writing: %s\n",
						 db_path, g_strerror(errno));

		exit(1);
	}

	load_db();

	g_free(db_dir);
	g_free(db_path);

	/* 10 minutes */
	db_file_timeout = g_timeout_add(FLUSH_TIMEOUT, timeout_cb, NULL);
}

void
galagod_person_db_uninit(void)
{
	galagod_person_db_flush();

	g_source_remove(db_file_timeout);
	db_file_timeout = 0;

	g_hash_table_destroy(db_table);
	db_table = NULL;

	if (db_fp != NULL)
	{
		fclose(db_fp);
		db_fp = NULL;
	}
}
