/**
 * @file person.c Galago Person message processor
 *
 * Copyright (C) 2004-2005 Christian Hammond.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA  02111-1307  USA
 */
#include "filter.h"
#include "meta-person.h"
#include "utils.h"

static void
_galagod_dbus_person_emit(GalagoPerson *person, const char *signal_name)
{
	DBusMessage *message;

	galago_return_if_fail(person != NULL);

	message = dbus_message_new_signal(galago_object_get_dbus_path(person),
									  GALAGO_DBUS_PERSON_INTERFACE,
									  signal_name);
	dbus_connection_send(galago_core_get_dbus_conn(), message, NULL);
	dbus_message_unref(message);
}

static void
_galagod_dbus_person_emit_with_object(GalagoPerson *person,
									  const char *signal_name, void *obj)
{
	DBusMessage *message;
	DBusMessageIter iter;

	galago_return_if_fail(person != NULL);
	galago_return_if_fail(obj != NULL && GALAGO_IS_OBJECT(obj));

	message = dbus_message_new_signal(galago_object_get_dbus_path(person),
									  GALAGO_DBUS_PERSON_INTERFACE,
									  signal_name);

	dbus_message_iter_init_append(message, &iter);
	galago_dbus_message_iter_append_object(&iter, obj);

	dbus_connection_send(galago_core_get_dbus_conn(), message, NULL);
	dbus_message_unref(message);
}

static DBusHandlerResult
person_set_property(DBusConnection *dbus_conn, DBusMessage *message,
					GalagodCoCo *coco, GalagoPerson *person)
{
	GalagodMetaPerson *meta_person;
	GalagoPerson *main_person;
	char *name;
	DBusMessage *signal;
	DBusMessageIter iter, signal_iter;

	meta_person = GALAGOD_META_PERSON(person);
	main_person = galagod_meta_person_get_person(meta_person);

	dbus_message_iter_init(message, &iter);

	signal = dbus_message_new_signal(galago_object_get_dbus_path(main_person),
									 GALAGO_DBUS_PERSON_INTERFACE,
									 "PropertySet");
	dbus_message_iter_init_append(signal, &signal_iter);

	galago_dbus_message_iter_get_string(&iter, name);
	dbus_message_iter_next(&iter);

	galago_dbus_message_iter_append_string(&signal_iter, name);

	if (dbus_message_iter_get_arg_type(&iter) == DBUS_TYPE_STRING)
	{
		char *value;
		galago_dbus_message_iter_get_string(&iter, value);
		galago_person_set_property_string(main_person, name, value);
		galago_dbus_message_iter_append_string(&signal_iter, value);
#if !GALAGO_CHECK_DBUS_VERSION(0, 30)
		dbus_free(value);
#endif
	}
	else if (dbus_message_iter_get_arg_type(&iter) == DBUS_TYPE_BOOLEAN)
	{
		dbus_bool_t value;
		galago_dbus_message_iter_get_boolean(&iter, value);
		galago_person_set_property_bool(main_person, name, value);
		galago_dbus_message_iter_append_boolean(&signal_iter, value);
	}
	else if (dbus_message_iter_get_arg_type(&iter) == DBUS_TYPE_UINT32)
	{
		dbus_uint32_t value;
		galago_dbus_message_iter_get_uint32(&iter, value);
		galago_person_set_property_uint32(main_person, name, value);
		galago_dbus_message_iter_append_uint32(&signal_iter, value);
	}

#if !GALAGO_CHECK_DBUS_VERSION(0, 30)
	dbus_free(name);
#endif

	dbus_connection_send(dbus_conn, signal, NULL);
	dbus_message_unref(signal);

	galagod_dbus_send_empty_reply(dbus_conn, message);

	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult
person_remove_property(DBusConnection *dbus_conn, DBusMessage *message,
					   GalagodCoCo *coco, GalagoPerson *person)
{
	GalagodMetaPerson *meta_person;
	GalagoPerson *main_person;
	char *name;
	DBusMessage *signal;
	DBusMessageIter iter, signal_iter;

	dbus_message_iter_init(message, &iter);
	galago_dbus_message_iter_get_string(&iter, name);

	meta_person = GALAGOD_META_PERSON(person);
	main_person = galagod_meta_person_get_person(meta_person);

	galago_person_remove_property(main_person, name);

	galagod_dbus_send_empty_reply(dbus_conn, message);

	/* Emit the signal */
	signal = dbus_message_new_signal(galago_object_get_dbus_path(main_person),
									 GALAGO_DBUS_PERSON_INTERFACE,
									 "PropertyUnset");
	dbus_message_iter_init_append(signal, &signal_iter);
	galago_dbus_message_iter_append_string(&signal_iter, name);
	dbus_connection_send(dbus_conn, signal, NULL);
	dbus_message_unref(signal);

#if !GALAGO_CHECK_DBUS_VERSION(0, 30)
	dbus_free(name);
#endif

	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult
person_get_property(DBusConnection *dbus_conn, DBusMessage *message,
					GalagodCoCo *coco, GalagoPerson *person)
{
	GalagodMetaPerson *meta_person;
	GalagoPerson *main_person;
	const GalagoValue *value;
	char *name;
	DBusMessageIter iter;
	DBusMessage *reply = NULL;

	dbus_message_iter_init(message, &iter);
	galago_dbus_message_iter_get_string(&iter, name);

	meta_person = GALAGOD_META_PERSON(person);
	main_person = galagod_meta_person_get_person(meta_person);

	value = galago_person_get_property(main_person, name);

	if (value == NULL)
	{
		/* TODO: Use dbus_message_new_error_printf when moving to 0.3x */
		char *buffer = g_strdup_printf("The property \"%s\" is invalid.",
									   name);

		reply = dbus_message_new_error(message,
			GALAGO_DBUS_ERROR_INVALID_PROPERTY,
			buffer);

		g_free(buffer);
	}
	else
	{
		GalagoType type = galago_value_get_type(value);

		reply = dbus_message_new_method_return(message);
		dbus_message_iter_init_append(reply, &iter);

		if (type == GALAGO_TYPE_STRING)
		{
			const char *str = galago_value_get_string(value);
			galago_dbus_message_iter_append_string(&iter, str);
		}
		else if (type == GALAGO_TYPE_BOOLEAN)
		{
			galago_bool b = galago_value_get_boolean(value);
			galago_dbus_message_iter_append_boolean(&iter, b);
		}
		else if (type == GALAGO_TYPE_UINT)
		{
			dbus_uint32_t i = galago_value_get_uint(value);
			galago_dbus_message_iter_append_uint32(&iter, i);
		}
		else
		{
			/* This should never happen. */
			galago_log_fatal("Unknown property type %d for property %s\n",
							 type, name);
		}
	}

	if (reply != NULL)
	{
		dbus_connection_send(dbus_conn, reply, NULL);
		dbus_message_unref(reply);
	}

#if !GALAGO_CHECK_DBUS_VERSION(0, 30)
	dbus_free(name);
#endif

	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult
person_get_account(DBusConnection *dbus_conn, DBusMessage *message,
				   GalagodCoCo *coco, GalagoPerson *person)
{
	GalagodMetaPerson *meta_person;
	GalagoPerson *main_person;
	GalagoAccount *account;
	GalagoService *service;
	char *username;
	DBusMessageIter iter;
	DBusMessage *reply;

	dbus_message_iter_init(message, &iter);
	service = galago_dbus_message_iter_get_object(&iter, GALAGO_CLASS_SERVICE);
	dbus_message_iter_next(&iter);
	galago_dbus_message_iter_get_string(&iter, username);

	meta_person = GALAGOD_META_PERSON(person);
	main_person = galagod_meta_person_get_person(meta_person);

	account = galago_person_get_account(main_person, service, username, FALSE);

	if (account == NULL)
	{
		reply = galagod_object_not_found_error_new(message,
			"The account '%s' on person '%s' could not be found.",
			username, galago_person_get_id(main_person));
	}
	else
	{
		reply = dbus_message_new_method_return(message);
		dbus_message_iter_init_append(reply, &iter);
		galago_dbus_message_iter_append_object(&iter, account);
	}

#if !GALAGO_CHECK_DBUS_VERSION(0, 30)
	dbus_free(username);
#endif

	dbus_connection_send(dbus_conn, reply, NULL);
	dbus_message_unref(reply);

	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult
person_get_accounts(DBusConnection *dbus_conn, DBusMessage *message,
					GalagodCoCo *coco, GalagoPerson *person)
{
	GalagodMetaPerson *meta_person;
	GalagoPerson *main_person;
	const GalagoList *list;
	DBusMessageIter iter;
	DBusMessage *reply;

	meta_person = GALAGOD_META_PERSON(person);
	main_person = galagod_meta_person_get_person(meta_person);

	list = galago_person_get_accounts(main_person, FALSE);

	reply = dbus_message_new_method_return(message);
	dbus_message_iter_init_append(reply, &iter);

	galago_dbus_message_iter_append_object_list(&iter, list);

	dbus_connection_send(dbus_conn, reply, NULL);
	dbus_message_unref(reply);

	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult
person_get_accounts_count(DBusConnection *dbus_conn, DBusMessage *message,
						  GalagodCoCo *coco, GalagoPerson *person)
{
	GalagodMetaPerson *meta_person;
	GalagoPerson *main_person;
	DBusMessageIter iter;
	DBusMessage *reply;
	size_t count;

	meta_person = GALAGOD_META_PERSON(person);
	main_person = galagod_meta_person_get_person(meta_person);

	count = galago_list_get_count(
		galago_person_get_accounts(main_person, FALSE));

	reply = dbus_message_new_method_return(message);
	dbus_message_iter_init_append(reply, &iter);
	galago_dbus_message_iter_append_uint32(&iter, count);
	dbus_connection_send(dbus_conn, reply, NULL);
	dbus_message_unref(reply);

	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult
set_unset_photo(DBusConnection *dbus_conn, DBusMessage *message,
				GalagodCoCo *coco, GalagoPerson *person, GalagoPhoto *photo)
{
	GalagodMetaPerson *meta_person;
	GalagoPerson *main_person;

	meta_person = GALAGOD_META_PERSON(person);
	main_person = galagod_meta_person_get_person(meta_person);

	if (photo != galago_person_get_photo(main_person, FALSE))
	{
		galago_person_set_photo(main_person, photo);

		galagod_dbus_send_empty_reply(dbus_conn, message);

		if (photo == NULL)
			_galagod_dbus_person_emit(main_person, "PhotoUnset");
		else
		{
			_galagod_dbus_person_emit_with_object(main_person, "PhotoSet",
												  photo);
		}
	}

	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult
person_set_photo(DBusConnection *dbus_conn, DBusMessage *message,
				 GalagodCoCo *coco, GalagoPerson *person)
{
	GalagoPhoto *photo;
	DBusMessageIter iter;

	dbus_message_iter_init(message, &iter);
	photo = galago_dbus_message_iter_get_object(&iter, GALAGO_CLASS_PHOTO);

	return set_unset_photo(dbus_conn, message, coco, person, photo);
}

static DBusHandlerResult
person_unset_photo(DBusConnection *dbus_conn, DBusMessage *message,
				   GalagodCoCo *coco, GalagoPerson *person)
{
	return set_unset_photo(dbus_conn, message, coco, person, NULL);
}

static DBusHandlerResult
person_get_photo(DBusConnection *dbus_conn, DBusMessage *message,
				 GalagodCoCo *coco, GalagoPerson *person)
{
	GalagodMetaPerson *meta_person;
	GalagoPerson *main_person;
	GalagoPhoto *photo;
	DBusMessage *reply;
	DBusMessageIter iter;

	meta_person = GALAGOD_META_PERSON(person);
	main_person = galagod_meta_person_get_person(meta_person);

	photo = galago_person_get_photo(main_person, FALSE);

	if (photo == NULL)
	{
		reply = galagod_object_not_found_error_new(message,
			"The photo for person '%s' could not be found.",
			galago_person_get_id(main_person));
	}
	else
	{
		reply = dbus_message_new_method_return(message);
		dbus_message_iter_init_append(reply, &iter);
		galago_dbus_message_iter_append_object(&iter, photo);
	}

	dbus_connection_send(dbus_conn, reply, NULL);
	dbus_message_unref(reply);

	return DBUS_HANDLER_RESULT_HANDLED;
}

static const GalagodCommand commands[] =
{
	GALAGOD_COMMAND("SetProperty",      person_set_property),
	GALAGOD_COMMAND("RemoveProperty",   person_remove_property),
	GALAGOD_COMMAND("GetProperty",      person_get_property),
	GALAGOD_COMMAND("GetAccount",       person_get_account),
	GALAGOD_COMMAND("GetAccounts",      person_get_accounts),
	GALAGOD_COMMAND("GetAccountsCount", person_get_accounts_count),
	GALAGOD_COMMAND("SetPhoto",         person_set_photo),
	GALAGOD_COMMAND("GetPhoto",         person_get_photo),
	GALAGOD_COMMAND("UnsetPhoto",       person_unset_photo),
	GALAGOD_COMMAND_LAST
};

void
galagod_person_add_filter(GalagoPerson *person)
{
	galagod_filters_add(person, commands, GALAGO_DBUS_PERSON_INTERFACE);
}

void
galagod_person_remove_filter(GalagoPerson *person)
{
	galagod_filters_remove(person);
}
