
/**
 * @file utils.c Utility functions
 *
 * Copyright (C) 2004-2005 Christian Hammond.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA  02111-1307  USA
 */
#include "galagod.h"
#include "utils.h"
#include "service-list.h"
#include "person-list.h"
#include "meta-service.h"
#include "meta-person.h"
#include "meta-presence.h"

void
galagod_dbus_send_empty_reply(DBusConnection *dbus_conn, DBusMessage *message)
{
	DBusMessage *reply;

	galago_return_if_fail(dbus_conn != NULL);
	galago_return_if_fail(message   != NULL);

	if (dbus_message_get_no_reply(message))
		return;

	reply = dbus_message_new_method_return(message);
	dbus_connection_send(dbus_conn, reply, NULL);
	dbus_message_unref(reply);
}

static void
dump_meta_account(GalagodMetaAccount *meta_account)
{
	const GalagoList *accounts, *l;

	accounts = galagod_meta_account_get_accounts(meta_account);

	galago_log_misc("    meta_account %p \"%s\" (count %zu)\n",
					meta_account,
					galago_account_get_username(
						galagod_meta_account_get_account(meta_account)),
					galago_list_get_count(accounts));

	for (l = accounts; l != NULL; l = l->next)
	{
		GalagoAccount *account = (GalagoAccount *)l->data;
		GalagodMetaPresence *meta_presence;
		const GalagoList *presences, *l2;

		meta_presence =
			GALAGOD_META_PRESENCE(galago_account_get_presence(account, FALSE));

		presences = galagod_meta_presence_get_presences(meta_presence);

		galago_log_misc("      %p (presence %p, count %zu)\n",
						account, meta_presence,
						galago_list_get_count(presences));

		for (l2 = presences; l2 != NULL; l2 = l2->next)
		{
			GalagoPresence *presence = (GalagoPresence *)l2->data;

			galago_log_misc("        %p\n", presence);
		}
	}
}

void
galagod_dump_data(void)
{
	const GalagoList *l, *l2;

	galago_log_misc("Services: (%zu)\n",
					galagod_services_get_meta_service_count());

	for (l = galagod_services_get_meta_services();
		 l != NULL;
		 l = l->next)
	{
		const GalagoList *services;
		GalagodMetaService *meta_service = (GalagodMetaService *)l->data;
		GalagodAccountList *account_list;

		services = galagod_meta_service_get_services(meta_service);

		galago_log_misc("  meta_service %p \"%s\" (count %zu)\n",
						meta_service,
						galago_service_get_name(
							galagod_meta_service_get_service(meta_service)),
						galago_list_get_count(services));

		for (l2 = services; l2 != NULL; l2 = l2->next)
		{
			GalagoService *service = (GalagoService *)l2->data;

			galago_log_misc("    %p\n", service);
		}

		account_list = galagod_meta_service_get_account_list(meta_service);

		galago_log_misc("    Accounts: (%zu)\n",
						galagod_account_list_get_meta_account_count(account_list));

		for (l2 = galagod_account_list_get_meta_accounts(account_list);
			 l2 != NULL;
			 l2 = l2->next)
		{
			dump_meta_account((GalagodMetaAccount *)l2->data);
		}
	}

	galago_log_misc("People: (%zu)\n",
					galagod_people_get_meta_person_count());

	for (l = galagod_people_get_meta_people();
		 l != NULL;
		 l = l->next)
	{
		const GalagoList *people;
		GalagodMetaPerson *meta_person = (GalagodMetaPerson *)l->data;

		people = galagod_meta_person_get_people(meta_person);

		galago_log_misc("  meta_person %p \"%s\" (count %zu)\n",
						meta_person,
						galago_person_get_id(
							galagod_meta_person_get_person(meta_person)),
						galago_list_get_count(people));

		for (l2 = people; l2 != NULL; l2 = l2->next)
		{
			galago_log_misc("    %p\n", l2->data);
		}
	}
}

DBusMessage *
galagod_object_not_found_error_new(DBusMessage *message,
								   const char *str, ...)
{
	DBusMessage *err_msg;
	char *error_str = NULL;
	va_list args;

	galago_return_val_if_fail(message != NULL, NULL);

	if (str != NULL)
	{
		va_start(args, str);
		error_str = g_strdup_vprintf(str, args);
		va_end(args);
	}

	err_msg = dbus_message_new_error(message,
									 GALAGO_DBUS_ERROR_OBJECT_NOT_FOUND,
									 error_str);

	if (error_str != NULL)
		g_free(error_str);

	return err_msg;
}
